(define-derived-mode slime-profile-browser-mode fundamental-mode
  "slprof"
  "Mode for browsing profiler data\
\\<slime-compiler-notes-mode-map>\
\\{slime-compiler-notes-mode-map}"
  (setq buffer-read-only t)
  (let ((inhibit-read-only t))
    (erase-buffer)
    (insert (format "%4s %-54s %6s %6s %6s\n"
                    "Rank"
                    "Name"
                    "Self%"
                    "Cumul%"
                    "Total%"))
    (dolist (data graph)
      (slime-profile-browser-insert-line data 54)))
  (goto-line 2))

(slime-define-keys slime-profile-browser-mode-map
  ("q" 'slime-temp-buffer-quit)
  ("d" 'slime-profile-browser-disassemble-function)
  ("g" 'slime-profile-browser-go-to)
  ("v" 'slime-profile-browser-view-source)
  ((kbd "RET") 'slime-profile-browser-toggle))

;; Start / stop profiling

(defun slime-profile-start ()
  (interactive)
  (slime-eval `(swank-backend::swank-profile-start)))

(defun slime-profile-stop ()
  (interactive)
  (slime-eval `(swank-backend::swank-profile-stop)))

;; Reporting

(defun slime-profile-browser ()
  (interactive)
  (lexical-let ((buffer (slime-profile-browser-get-buffer)))
    (slime-eval-async `(swank-backend::swank-profile-get-call-graph)
                      (lambda (graph)
                        (with-current-buffer buffer
                          (switch-to-buffer buffer)
                          (slime-profile-browser-mode))))))

(defun slime-profile-browser-get-buffer ()
  (get-buffer-create "*slime-profile-browser*"))

(defun slime-profile-browser-insert-line (data name-length)
  (destructuring-bind (index name self cumul total)
      data
    (if index
        (insert (format "%-4d " index))
        (insert "     "))
    (slime-insert-propertized
     (slime-profile-browser-name-properties)
     (format (format "%%-%ds " name-length)
             (abbreviate-name name name-length)))
    (insert (format "%6.2f " self))
    (when cumul
      (insert (format "%6.2f " cumul))
      (when total
        (insert (format "%6.2f" total))))
    (when index
      (slime-profile-browser-add-line-text-properties
       `(profile-index ,index expanded nil)))
    (insert "\n")))

(defun abbreviate-name (name max-length)
  (lexical-let ((length (min (length name) max-length)))
    (subseq name 0 length)))

;; Expanding / collapsing

(defun slime-profile-browser-toggle ()
  (interactive)
  (let ((index (get-text-property (point) 'profile-index)))
    (when index
      (save-excursion
        (if (slime-profile-browser-line-expanded-p)
            (slime-profile-browser-collapse)
            (slime-profile-browser-expand))))))

(defun slime-profile-browser-collapse ()
  (let ((inhibit-read-only t))
    (slime-profile-browser-add-line-text-properties '(expanded nil))
    (forward-line)
    (loop until (or (eobp)
                    (get-text-property (point) 'profile-index))
          do
          (delete-region (point-at-bol) (point-at-eol))
          (unless (eobp)
            (delete-char 1)))))

(defun slime-profile-browser-expand ()
  (lexical-let* ((buffer (current-buffer))
                 (point (point))
                 (index (get-text-property point 'profile-index)))
    (slime-eval-async `(swank-backend::swank-profile-expand-node ,index)
                      (lambda (data)
                        (with-current-buffer buffer
                          (save-excursion 
                            (destructuring-bind (&key callers calls)
                                data
                              (slime-profile-browser-add-expansion callers
                                                                   "Callers"
                                                                   0)
                              (slime-profile-browser-add-expansion calls
                                                                   "Calls"
                                                                   0))))))))

(defun slime-profile-browser-add-expansion (data type nesting)
  (when data
    (let ((inhibit-read-only t))
      (slime-profile-browser-add-line-text-properties '(expanded t))
      (end-of-line)
      (insert (format "\n     %s" type))
      (dolist (node data)
        (destructuring-bind (index name cumul) node
          (insert (format (format "\n%%%ds" (+ 7 (* 2 nesting))) ""))
          (slime-insert-propertized
           (slime-profile-browser-name-properties)
           (let ((len (- 59 (* 2 nesting))))
             (format (format "%%-%ds " len)
                     (abbreviate-name name len))))
          (slime-profile-browser-add-line-text-properties
           `(profile-sub-index ,index))
          (insert (format "%6.2f" cumul)))))))

(defun slime-profile-browser-line-expanded-p ()
  (get-text-property (point) 'expanded))

(defun slime-profile-browser-add-line-text-properties (properties)
  (add-text-properties (point-at-bol)
                       (point-at-eol)
                       properties))

(defun slime-profile-browser-name-properties ()
  '(face sldb-restart-number-face))

;; "Go to function"

(defun slime-profile-browser-go-to ()                                           
  (interactive)
  (let ((sub-index (get-text-property (point) 'profile-sub-index)))
    (when sub-index
      (let ((pos (text-property-any
                  (point-min) (point-max) 'profile-index sub-index)))
        (when pos (goto-char pos))))))

;; Disassembly

(defun slime-profile-browser-disassemble-function ()
  (interactive)
  (let ((index (or (get-text-property (point) 'profile-index)
                   (get-text-property (point) 'profile-sub-index))))
    (when index
      (slime-eval-describe `(swank-backend::swank-profile-disassemble
                             ,index)))))

;; View source

(defun slime-profile-browser-view-source ()
  (interactive)
  (let ((index (or (get-text-property (point) 'profile-index)
                   (get-text-property (point) 'profile-sub-index))))
    (when index
      (slime-eval-async
       `(swank-backend::swank-profile-source-location ,index)
       (lambda (source-location)
         (destructure-case source-location
           ((:error message)
            (message "%s" message)
            (ding))
           (t
            (slime-show-source-location source-location))))))))
